@@ -59,29 +59,31 @@ gdip_brush_setup (GpGraphics *graphics, GpBrush *brush)
 	}
 }
 
-GpStatus
-gdip_brush_destroy (GpBrush *brush)
-{
-	
-	/* do nothing, because we are not allocating any memory here. */
-	return Ok;
-}
-
+/* coverity[+alloc : arg-*1] */
 GpStatus 
 GdipCloneBrush (GpBrush *brush, GpBrush **clonedBrush)
 {
+	if (!brush || !clonedBrush)
+		return InvalidParameter;
+
 	return brush->vtable->clone_brush (brush, clonedBrush);
 }
 
 GpStatus 
 GdipDeleteBrush (GpBrush *brush)
 {
+	if (!brush)
+		return InvalidParameter;
+
 	return brush->vtable->destroy (brush);
 }
 
 GpStatus
 GdipGetBrushType (GpBrush *brush, GpBrushType *type)
 {
+	if (!brush || !type)
+		return InvalidParameter;
+
         *type = brush->vtable->type;
         return Ok;
 }
