@@ -30,6 +30,7 @@
 
 
 /* Generic fonts families */
+static GStaticMutex generic = G_STATIC_MUTEX_INIT;
 static GpFontFamily* familySerif = NULL;
 static GpFontFamily* familySansSerif = NULL;
 static GpFontFamily* familyMonospace = NULL;
@@ -37,31 +38,27 @@ static int ref_familySerif = 0;
 static int ref_familySansSerif = 0;
 static int ref_familyMonospace = 0;
 
-extern cairo_font_face_t *
-_cairo_toy_font_face_create (const char *family, 
-			     cairo_font_slant_t   slant, 
-			     cairo_font_weight_t  weight);
-
 
 /* Family and collections font functions */
 
 void
 gdip_createFontFamily (GpFontFamily **family)
 {
-	GpFontFamily *result;
-
-	result = (GpFontFamily *) GdipAlloc (sizeof (GpFontFamily));
-	result->height = -1;
-	result->linespacing = -1;
-	result->celldescent = -1;
-	result->cellascent = -1;
-	result->pattern = NULL;
-	result->allocated = FALSE;
+	GpFontFamily *result = (GpFontFamily *) GdipAlloc (sizeof (GpFontFamily));
+	if (result) {
+		result->height = -1;
+		result->linespacing = -1;
+		result->celldescent = -1;
+		result->cellascent = -1;
+		result->pattern = NULL;
+		result->allocated = FALSE;
+	}
 	*family = result;
 }
 
 static GpFontCollection *system_fonts = NULL;
 
+// coverity[+alloc : arg-*0]
 GpStatus
 GdipNewInstalledFontCollection (GpFontCollection **font_collection)
 {	
@@ -90,14 +87,17 @@ GdipNewInstalledFontCollection (GpFontCollection **font_collection)
 		FcObjectSetDestroy (os);
     
 		system_fonts = (GpFontCollection *) GdipAlloc (sizeof (GpFontCollection));
-		system_fonts->fontset = col;
-		system_fonts->config = NULL;
+		if (system_fonts) {
+			system_fonts->fontset = col;
+			system_fonts->config = NULL;
+		}
 	}
 
 	*font_collection = system_fonts;
 	return Ok;
 }
 
+// coverity[+alloc : arg-*0]
 GpStatus
 GdipNewPrivateFontCollection (GpFontCollection **font_collection)
 {
@@ -107,13 +107,15 @@ GdipNewPrivateFontCollection (GpFontCollection **font_collection)
 		return InvalidParameter;
 
 	result = (GpFontCollection *) GdipAlloc (sizeof (GpFontCollection));
-	result->fontset = NULL;
-	result->config = FcConfigCreate ();
-    
+	if (result) {
+		result->fontset = NULL;
+		result->config = FcConfigCreate ();
+    	}
 	*font_collection = result;
 	return Ok;
 }
 
+// coverity[+free : arg-0]
 GpStatus
 GdipDeletePrivateFontCollection (GpFontCollection **font_collection)
 {
@@ -144,6 +146,8 @@ GdipPrivateAddFontFile (GpFontCollection *font_collection,  GDIPCONST WCHAR *fil
 		return InvalidParameter;
     
 	file = (unsigned char*) ucs2_to_utf8 ((const gunichar2 *)filename, -1);
+	if (!file)
+		return OutOfMemory;
 
 	FcConfigAppFontAddFile (font_collection->config, file);
     
@@ -158,25 +162,35 @@ GdipDeleteFontFamily (GpFontFamily *fontFamily)
 	bool delete = TRUE;
 	
 	if (!fontFamily)
-		return Ok;
-		
+		return InvalidParameter;
+
+	g_static_mutex_lock (&generic);
+	
 	if (fontFamily == familySerif) {
 		ref_familySerif--;
 		if (ref_familySerif)
 			delete = FALSE;
+		else
+			familySerif = NULL;
 	}
 	
 	if (fontFamily == familySansSerif) {
 		ref_familySansSerif--;
 		if (ref_familySansSerif)
 			delete = FALSE;
+		else
+			familySansSerif = NULL;
 	}	
 	
 	if (fontFamily == familyMonospace) {
 		ref_familyMonospace--;
 		if (ref_familyMonospace)
 			delete = FALSE;
+		else
+			familyMonospace = NULL;
 	}	
+
+	g_static_mutex_unlock (&generic);
 	
 	if (delete) {
 		if (fontFamily->allocated) {
@@ -243,67 +257,179 @@ GdipGetFontCollectionFamilyList (GpFontCollection *font_collection, int num_soug
 	return Ok;  
 }
 
+static GpStatus
+gdip_status_from_fontconfig (FcResult result)
+{
+	switch (result) {
+	case FcResultMatch:
+		return Ok;
+	case FcResultNoMatch:
+	case FcResultTypeMismatch:
+	case FcResultNoId:
+		return FontFamilyNotFound;
+	default:
+		return GenericError;
+	}
+}
+
+/* note: MUST be executed inside a lock because FcConfig isn't thread-safe */
+static FcPattern*
+create_pattern_from_name (char* name)
+{
+	FcValue val;
+	/* FcResult must be initialized because it's changed only in error conditions */
+	FcResult rlt = FcResultMatch;
+	FcPattern *full_pattern = NULL;
+	FcPattern *name_pattern = FcPatternCreate ();
+
+	if (!name_pattern)
+		return NULL;
+		
+	/* find the family we want */
+	val.type = FcTypeString;
+	val.u.s = (unsigned char*)name;
+	if (!FcPatternAdd (name_pattern, FC_FAMILY, val, TRUE)) {
+		FcPatternDestroy (name_pattern);
+		return NULL;
+	}
+
+	if (!FcConfigSubstitute (0, name_pattern, FcMatchPattern)) {
+		FcPatternDestroy (name_pattern);
+		return NULL;
+	}
+
+	FcDefaultSubstitute (name_pattern);                  
+		
+	full_pattern = FcFontMatch (0, name_pattern, &rlt);
+	if (gdip_status_from_fontconfig (rlt) == Ok) {
+		if (full_pattern == NULL) {
+			full_pattern = name_pattern;
+		} else {
+			FcPatternDestroy (name_pattern);
+		}
+	} else {
+		FcPatternDestroy (name_pattern);
+		if (full_pattern) {
+			FcPatternDestroy (full_pattern);
+			full_pattern = NULL;
+		}
+	}
+
+	return full_pattern;
+}
+
+static GStaticMutex patterns_mutex = G_STATIC_MUTEX_INIT;
+static GHashTable *patterns_hashtable;
+
+static GpStatus
+create_fontfamily_from_name (char* name, GpFontFamily **fontFamily)
+{
+	GpStatus status = FontFamilyNotFound;
+	GpFontFamily *ff = NULL;
+	FcPattern *pat = NULL;
+
+	g_static_mutex_lock (&patterns_mutex);
+
+	if (patterns_hashtable) {
+		pat = (FcPattern*) g_hash_table_lookup (patterns_hashtable, name);
+	} else {
+		patterns_hashtable = g_hash_table_new (g_str_hash, g_str_equal);
+	}
+
+	if (!pat) {
+		pat = create_pattern_from_name (name);
+		if (pat) {
+			/* create the pattern and store it for further usage */
+			g_hash_table_insert (patterns_hashtable, g_strdup (name), pat);
+		}
+	}
+
+	if (pat) {
+		gdip_createFontFamily (&ff);
+		if (ff) {
+			ff->pattern = pat;
+			ff->allocated = FALSE;
+			status = Ok;
+		} else 
+			status = OutOfMemory;
+	}
+
+	*fontFamily = ff;
+	g_static_mutex_unlock (&patterns_mutex);
+	return status;
+}
+
+static
+gboolean free_cached_pattern (gpointer key, gpointer value, gpointer user)
+{
+	g_free (value);
+	FcPatternDestroy ((FcPattern*) value);
+	return TRUE;
+}
+
+void
+gdip_font_clear_pattern_cache (void)
+{
+	g_static_mutex_lock (&patterns_mutex);
+	g_hash_table_foreach_remove (patterns_hashtable, free_cached_pattern, NULL);
+	g_hash_table_destroy (patterns_hashtable);
+	g_static_mutex_unlock (&patterns_mutex);
+}
+
+static GpStatus
+create_fontfamily_from_collection (char* name, GpFontCollection *font_collection, GpFontFamily **fontFamily)
+{
+	/* note: fontset can be NULL when we supply an empty private collection */
+	if (font_collection->fontset) {
+		int i;
+		FcChar8 *str;
+		FcPattern **gpfam = font_collection->fontset->fonts;
+    
+		for (i=0; i < font_collection->fontset->nfont; gpfam++, i++) {
+			FcResult rlt = FcPatternGetString (*gpfam, FC_FAMILY, 0, &str);
+			GpStatus status = gdip_status_from_fontconfig (rlt);
+			if (status != Ok)
+				return status;
+
+			if (strcmp ((char *)name, (const char *)str) == 0) {
+				gdip_createFontFamily (fontFamily);
+				(*fontFamily)->pattern = *gpfam;
+				(*fontFamily)->allocated = FALSE;
+				return Ok;
+			}
+		}
+	}
+	return FontFamilyNotFound;
+}
+
+// coverity[+alloc : arg-*2]
 GpStatus
 GdipCreateFontFamilyFromName (GDIPCONST WCHAR *name, GpFontCollection *font_collection, GpFontFamily **fontFamily)
 {
-	unsigned char *string;
-	FcPattern **gpfam;
-	FcChar8 *str;
-	int i;
+	GpStatus status;
+	char *string;
 	
 	if (!name || !fontFamily)
 		return InvalidParameter;
 
-	string = (unsigned char*)ucs2_to_utf8 ((const gunichar2 *)name, -1);
+	string = (char*)ucs2_to_utf8 ((const gunichar2 *)name, -1);
+	if (!string)
+		return OutOfMemory;
 
-	if (!font_collection) {
-		FcPattern *pat = FcPatternCreate ();
-		FcResult rlt;
-		
-		/* find the family we want */
-		FcValue val;
-                                   
-		val.type = FcTypeString;
-		val.u.s = string;
-		FcPatternAdd (pat, FC_FAMILY, val, TRUE);
-
-		FcConfigSubstitute (0, pat, FcMatchPattern);
-		FcDefaultSubstitute (pat);                  
-		
-		gdip_createFontFamily (fontFamily);
-		(*fontFamily)->pattern =  FcFontMatch (0, pat, &rlt);
-		(*fontFamily)->allocated = TRUE;
-		GdipFree (string);
-		if ((*fontFamily)->pattern == NULL) {
-			(*fontFamily)->pattern = pat;
-		} else {
-			FcPatternDestroy (pat);			
-		}
-		
-		return Ok;
-	}
-
-	gpfam = font_collection->fontset->fonts;
-    
-	for (i=0; i < font_collection->fontset->nfont; gpfam++, i++){
-		FcPatternGetString (*gpfam, FC_FAMILY, 0, &str);
-
-		if (strcmp ((char *)string, (const char *)str)==0) {
-			gdip_createFontFamily (fontFamily);
-			(*fontFamily)->pattern = *gpfam;
-			(*fontFamily)->allocated = FALSE;
-			GdipFree (string);
-			return Ok;
-		}
+	if (font_collection) {
+		status = create_fontfamily_from_collection (string, font_collection, fontFamily);
+	} else {
+		status = create_fontfamily_from_name (string, fontFamily);
 	}
 
 	GdipFree (string);
-	return FontFamilyNotFound;
+	return status;
 }
 
 GpStatus
 GdipGetFamilyName (GDIPCONST GpFontFamily *family, WCHAR name[LF_FACESIZE], int language)
-{                
+{
+	GpStatus status;
 	FcChar8 *fc_str;
 	FcResult r;
 	
@@ -311,49 +437,81 @@ GdipGetFamilyName (GDIPCONST GpFontFamily *family, WCHAR name[LF_FACESIZE], int
 		return InvalidParameter;
 
 	r = FcPatternGetString (family->pattern, FC_FAMILY, 0, &fc_str);
+	status = gdip_status_from_fontconfig (r);
+	if (status != Ok)
+		return status;
 
 	utf8_to_ucs2((const gchar *)fc_str, (gunichar2 *)name, LF_FACESIZE);
 	return Ok;
 }
 
-
+// coverity[+alloc : arg-*0]
 GpStatus
 GdipGetGenericFontFamilySansSerif (GpFontFamily **nativeFamily)
 {
 	const WCHAR MSSansSerif[] = {'M','S',' ','S','a','n','s',' ', 'S','e','r','i','f', 0};
+	GpStatus status = Ok;
 	
-	if (ref_familySansSerif == 0) 
-		GdipCreateFontFamilyFromName (MSSansSerif, NULL, &familySansSerif);    
-	
-	ref_familySansSerif++;
+	g_static_mutex_lock (&generic);
+
+	if (ref_familySansSerif == 0)
+		status = GdipCreateFontFamilyFromName (MSSansSerif, NULL, &familySansSerif);
+
+	if (status == Ok)
+		ref_familySansSerif++;
+	else
+		familySansSerif = NULL;
+
+	g_static_mutex_unlock (&generic);
+
 	*nativeFamily = familySansSerif;    
-	return Ok;
+	return status;
 }
 
+// coverity[+alloc : arg-*0]
 GpStatus
 GdipGetGenericFontFamilySerif (GpFontFamily **nativeFamily)
 {
 	const WCHAR Serif[] = {'S','e','r','i','f', 0};
+	GpStatus status = Ok;
 	
+	g_static_mutex_lock (&generic);
+
 	if (ref_familySerif == 0)
-		GdipCreateFontFamilyFromName (Serif, NULL, &familySerif);
-	
-	ref_familySerif++;	
-	*nativeFamily = familySerif;    
-	return Ok;
+		status = GdipCreateFontFamilyFromName (Serif, NULL, &familySerif);
+
+	if (status == Ok)
+		ref_familySerif++;
+	else
+		familySerif = NULL;
+
+	g_static_mutex_unlock (&generic);
+
+	*nativeFamily = familySerif;
+	return status;
 }
 
+// coverity[+alloc : arg-*0]
 GpStatus
 GdipGetGenericFontFamilyMonospace (GpFontFamily **nativeFamily)
 {
 	const WCHAR Serif[] = {'S','e','r','i','f', 0};
+	GpStatus status = Ok;
 	
+	g_static_mutex_lock (&generic);
+
 	if (ref_familyMonospace == 0)
-		GdipCreateFontFamilyFromName (Serif, NULL, &familyMonospace);    
-		
-	ref_familyMonospace++;
+		status = GdipCreateFontFamilyFromName (Serif, NULL, &familyMonospace);
+
+	if (status == Ok)
+		ref_familyMonospace++;
+	else
+		familyMonospace = NULL;
+
+	g_static_mutex_unlock (&generic);
+
 	*nativeFamily = familyMonospace;    
-	return Ok;
+	return status;
 }
 
 FT_Face
@@ -371,6 +529,15 @@ gdip_cairo_ft_font_lock_face (cairo_font_face_t *cairofnt, cairo_scaled_font_t *
 					      &matrix1,
 					      &matrix2,
 					      options);
+	/* a missing fonts.conf will resuls in a NULL *scaled_ft (#78237) */
+	if (!*scaled_ft) {
+		static int flag = 0;
+		if (flag == 0) {
+			g_warning ("couldn't lock the font face. this may be due to a missing fonts.conf on the system.");
+			flag = 1;
+		}
+		return NULL;
+	}
 
 	face = cairo_ft_scaled_font_lock_face (*scaled_ft);
 
@@ -407,16 +574,15 @@ GdipGetEmHeight (GDIPCONST GpFontFamily *family, GpFontStyle style, short *EmHei
 		TT_VertHeader *pVert;
 
 		face = gdip_cairo_ft_font_lock_face (font->cairofnt, &scaled_ft);
-
-		pVert = FT_Get_Sfnt_Table(face, ft_sfnt_vhea);
-		if (pVert) {
-			rslt = pVert->yMax_Extent;
-		} else if (face) {
-			rslt = face->units_per_EM;
-		} else {
-			rslt = 0;
+		if (face) {
+			pVert = FT_Get_Sfnt_Table(face, ft_sfnt_vhea);
+			if (pVert) {
+				rslt = pVert->yMax_Extent;
+			} else if (face) {
+				rslt = face->units_per_EM;
+			}
+			gdip_cairo_ft_font_unlock_face (scaled_ft);
 		}
-		gdip_cairo_ft_font_unlock_face(scaled_ft);
 		GdipDeleteFont (font);
 	}
 
@@ -447,13 +613,13 @@ GdipGetCellAscent (GDIPCONST GpFontFamily *family, GpFontStyle style, short *Cel
 		TT_HoriHeader *pHori;
 
 		face = gdip_cairo_ft_font_lock_face (font->cairofnt, &scaled_ft);
+		if (face) {
+			pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
+			if (pHori)
+				rslt = pHori->Ascender;
 
-		pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
-
-		if (pHori)
-			rslt = pHori->Ascender;
-
-		gdip_cairo_ft_font_unlock_face(scaled_ft);
+			gdip_cairo_ft_font_unlock_face (scaled_ft);
+		}
 		GdipDeleteFont (font);
 	}
 
@@ -484,13 +650,13 @@ GdipGetCellDescent (GDIPCONST GpFontFamily *family, GpFontStyle style, short *Ce
 		TT_HoriHeader* pHori;
 
 		face = gdip_cairo_ft_font_lock_face (font->cairofnt, &scaled_ft);
+		if (face) {
+			pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
+			if (pHori)
+				rslt = -pHori->Descender;
 
-		pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
-
-		if (pHori)
-			rslt = -pHori->Descender;
-
-		gdip_cairo_ft_font_unlock_face(scaled_ft);
+			gdip_cairo_ft_font_unlock_face (scaled_ft);
+		}
 		GdipDeleteFont (font);
 	}
 
@@ -521,17 +687,16 @@ GdipGetLineSpacing (GDIPCONST GpFontFamily *family, GpFontStyle style, short *Li
 		TT_HoriHeader *pHori;
 
 		face = gdip_cairo_ft_font_lock_face (font->cairofnt, &scaled_ft);
+		if (face) {
+			pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
+			if (pHori) {
+				rslt = pHori->Ascender + (-pHori->Descender) + pHori->Line_Gap;
+			} else if (face) {
+				rslt = face->height;
+			}
 
-		pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
-		if (pHori) {
-			rslt = pHori->Ascender + (-pHori->Descender) + pHori->Line_Gap;
-		} else if (face) {
-			rslt = face->height;
-		} else {
-			rslt = 0;
+			gdip_cairo_ft_font_unlock_face(scaled_ft);
 		}
-
-		gdip_cairo_ft_font_unlock_face(scaled_ft);
 		GdipDeleteFont (font);
 	}
 
@@ -551,10 +716,29 @@ GdipIsStyleAvailable (GDIPCONST GpFontFamily *family, int style, BOOL *IsStyleAv
 }
 
 /* Font functions */
+cairo_font_face_t *
+gdip_face_create (const char *family, 
+			cairo_font_slant_t   slant, 
+			cairo_font_weight_t  weight,
+			cairo_t** ct)
+{
+	cairo_surface_t *surface;
+	cairo_font_face_t *face;
 
+	surface = cairo_image_surface_create_for_data ((unsigned char *)NULL, CAIRO_FORMAT_ARGB32, 0, 0, 0);
+	*ct = cairo_create (surface);
+	cairo_select_font_face (*ct, (const char *) family, slant, weight);
+	face = cairo_get_font_face (*ct);
+	cairo_surface_destroy (surface);
+	return face;
+
+}
+
+// coverity[+alloc : arg-*4]
 GpStatus
 GdipCreateFont (GDIPCONST GpFontFamily* family, float emSize, GpFontStyle style, Unit unit,  GpFont **font)
 {
+	GpStatus status;
 	FcChar8* str;
 	FcResult r;
 	GpFont *result;
@@ -563,12 +747,15 @@ GdipCreateFont (GDIPCONST GpFontFamily* family, float emSize, GpFontStyle style,
 	float sizeInPixels;
 	cairo_font_face_t* cairofnt;
 	
-	if (!family || !font)
+	if (!family || !font || (unit == UnitDisplay))
 		return InvalidParameter;
 
 	r = FcPatternGetString (family->pattern, FC_FAMILY, 0, &str);
+	status = gdip_status_from_fontconfig (r);
+	if (status != Ok)
+		return status;
 	
-	gdip_unit_conversion (unit, UnitPixel, gdip_get_display_dpi(), gtMemoryBitmap, emSize, &sizeInPixels);
+	sizeInPixels = gdip_unit_conversion (unit, UnitPixel, gdip_get_display_dpi(), gtMemoryBitmap, emSize);
 		
 	result = (GpFont *) GdipAlloc (sizeof (GpFont));
 	result->sizeInPixels = sizeInPixels;
@@ -583,7 +770,7 @@ GdipCreateFont (GDIPCONST GpFontFamily* family, float emSize, GpFontStyle style,
 	else
 		slant = CAIRO_FONT_SLANT_NORMAL;
 
-	cairofnt = _cairo_toy_font_face_create ((const char*) str, slant, weight);
+	cairofnt = gdip_face_create ((const char*) str, slant, weight, &result->ct);
 
 	if (cairofnt == NULL) {
 		GdipFree(result);
@@ -625,7 +812,8 @@ GdipDeleteFont (GpFont* font)
 	if (!font)
 		return InvalidParameter;
 
-	cairo_font_face_destroy (font->cairofnt);
+	cairo_destroy (font->ct);
+
 	GdipFree ((void *)font->face);
 	GdipFree ((void *)font);
 	return Ok;	       
@@ -637,13 +825,22 @@ GdipCreateFontFromDC(void *hdc, GpFont **font)
 	return(NotImplemented);
 }
 
-static void
+static GpStatus
 gdip_logfont_from_font(GpFont *font, GpGraphics *graphics, void *lf, bool ucs2)
 {
 	LOGFONTA		*logFont;
 
+	if (!lf)
+		return InvalidParameter;
+
 	logFont = (LOGFONTA *)lf;
 
+	/* will be changed back to 1 inside System.Drawing */
+	logFont->lfCharSet = 0;
+
+	if (!font || !graphics)
+		return InvalidParameter;
+
 	logFont->lfHeight = -(font->sizeInPixels);
 	logFont->lfWidth = 0;
 	logFont->lfEscapement = 0;	// FIXME
@@ -672,42 +869,41 @@ gdip_logfont_from_font(GpFont *font, GpGraphics *graphics, void *lf, bool ucs2)
 		logFont->lfStrikeOut = 0;
 	}
 
-	logFont->lfCharSet = 1;	// DEFAULT_CHARSET
 	logFont->lfOutPrecision = 0;
 	logFont->lfClipPrecision = 0;
-	if (graphics != NULL) {
-		switch(graphics->text_mode) {
-			case TextRenderingHintSystemDefault: {
-				logFont->lfQuality = 0;
-				break;
-			}
+
+	switch (graphics->text_mode) {
+		case TextRenderingHintSystemDefault: {
+			logFont->lfQuality = 0;
+			break;
+		}
 	
-			case TextRenderingHintSingleBitPerPixelGridFit:
-			case TextRenderingHintSingleBitPerPixel:
-			case TextRenderingHintAntiAliasGridFit:
-			case TextRenderingHintAntiAlias: {
-				logFont->lfQuality = 3;	// ANTIALIASED_QUALITY;
-				break;
-			}
+		case TextRenderingHintSingleBitPerPixelGridFit:
+		case TextRenderingHintSingleBitPerPixel:
+		case TextRenderingHintAntiAliasGridFit:
+		case TextRenderingHintAntiAlias: {
+			logFont->lfQuality = 3;	// ANTIALIASED_QUALITY;
+			break;
+		}
 	
-			case TextRenderingHintClearTypeGridFit: {
-				logFont->lfQuality = 5;	// CLEARTYPE_QUALITY
-				break;
-			}
+		case TextRenderingHintClearTypeGridFit: {
+			logFont->lfQuality = 5;	// CLEARTYPE_QUALITY
+			break;
 		}
-	} else {
-		logFont->lfQuality = 0;
 	}
 
 	logFont->lfPitchAndFamily = 0;
 	if (ucs2) {
 		utf8_to_ucs2((const gchar *)font->face, (gunichar2 *)logFont->lfFaceName, LF_FACESIZE);
 	} else {
-		memcpy(logFont->lfFaceName, font->face, LF_FACESIZE);
-		logFont->lfFaceName[LF_FACESIZE - 1] = '\0';
+		int len = strlen ((char*)font->face);
+		memset (logFont->lfFaceName, 0, LF_FACESIZE);
+		memcpy (logFont->lfFaceName, font->face, len < LF_FACESIZE ? len : LF_FACESIZE - 1);
 	}
+	return Ok;
 }
 
+// coverity[+alloc : arg-*1]
 GpStatus
 GdipCreateFontFromHfontA(void *hfont, GpFont **font, void *lf)
 {
@@ -720,6 +916,9 @@ GdipCreateFontFromHfontA(void *hfont, GpFont **font, void *lf)
 	src_font = (GpFont *)hfont;
 
 	result = (GpFont *) GdipAlloc (sizeof (GpFont));
+	if (!result)
+		return OutOfMemory;
+
 	result->sizeInPixels = src_font->sizeInPixels;
 	result->style = src_font->style;
 	result->family = src_font->family;
@@ -746,7 +945,7 @@ GdipCreateFontFromHfontA(void *hfont, GpFont **font, void *lf)
 
 	memcpy(result->face, src_font->face, strlen((char *)src_font->face) + 1);
 
-	result->cairofnt = _cairo_toy_font_face_create ((const char*) src_font->face, slant, weight);
+	result->cairofnt = gdip_face_create ((const char*) src_font->face, slant, weight, &result->ct);
 
 	if (result->cairofnt == NULL) {
 		GdipFree(result);
@@ -755,31 +954,19 @@ GdipCreateFontFromHfontA(void *hfont, GpFont **font, void *lf)
 
 	*font = result;
 
-	gdip_logfont_from_font(result, NULL, lf, FALSE);
-
-	return Ok;
+	return gdip_logfont_from_font (result, NULL, lf, FALSE);
 }
 
 GpStatus
 GdipGetLogFontW(GpFont *font, GpGraphics *graphics, void *lf)
 {
-	if (!font || !lf) {
-		return(InvalidParameter);
-	}
-	gdip_logfont_from_font(font, graphics, lf, TRUE);
-	return Ok;
+	return gdip_logfont_from_font (font, graphics, lf, TRUE);
 }
 
 GpStatus
 GdipGetLogFontA(GpFont *font, GpGraphics *graphics, void *lf)
 {
-		
-	if (!font || !lf) {
-		return(InvalidParameter);
-	}
-	
-	gdip_logfont_from_font(font, graphics, lf, FALSE);
-	return Ok;
+	return gdip_logfont_from_font (font, graphics, lf, FALSE);
 }
 
 GpStatus
@@ -828,13 +1015,21 @@ gdip_create_font_from_logfont(void *hdc, void *lf, GpFont **font, bool ucs2)
 
 	if (ucs2) {
 		result->face = (unsigned char*) ucs2_to_utf8 ((const gunichar2 *)logfont->lfFaceName, -1);
+		if (!result->face){
+			GdipFree (result);
+			return OutOfMemory;
+		}
 	} else {
 		result->face = GdipAlloc(LF_FACESIZE);
+		if (!result->face){
+			GdipFree (result);
+			return OutOfMemory;
+		}
 		memcpy(result->face, logfont->lfFaceName, LF_FACESIZE);
 		result->face[LF_FACESIZE - 1] = '\0';
 	}
 
-	result->cairofnt = _cairo_toy_font_face_create ((const char *)result->face, slant, weight);
+	result->cairofnt = gdip_face_create ((const char *)result->face, slant, weight, &result->ct);
 	if (result->cairofnt == NULL) {
 		GdipFree(result);
 		return GenericError;
@@ -845,12 +1040,14 @@ gdip_create_font_from_logfont(void *hdc, void *lf, GpFont **font, bool ucs2)
 	return Ok;
 }
 
+// coverity[+alloc : arg-*2]
 GpStatus
 GdipCreateFontFromLogfontA(void *hdc, GDIPCONST LOGFONTA *logfont, GpFont **font)
 {
 	return gdip_create_font_from_logfont(hdc, (void *)logfont, font, FALSE);
 }
 
+// coverity[+alloc : arg-*2]
 GpStatus
 GdipCreateFontFromLogfontW(void *hdc, GDIPCONST LOGFONTW *logfont, GpFont **font)
 {
@@ -921,12 +1118,12 @@ GdipGetFontHeight (GDIPCONST GpFont *font, GDIPCONST GpGraphics *graphics, float
 		return InvalidParameter;
 
 	/* Operations in display dpi's */	
-	gdip_unit_conversion (font->unit, UnitPixel, gdip_get_display_dpi (), gtMemoryBitmap, font->emSize, &emSize);
+	emSize = gdip_unit_conversion (font->unit, UnitPixel, gdip_get_display_dpi (), gtMemoryBitmap, font->emSize);
 	GdipGetEmHeight (font->family, font->style, &emHeight);
 	GdipGetLineSpacing (font->family, font->style, &lineSpacing);
 
 	*height = lineSpacing * (emSize / emHeight);
-	gdip_unit_conversion (UnitPixel, graphics->page_unit, gdip_get_display_dpi (), graphics->type, *height, height);
+	*height = gdip_unit_conversion (UnitPixel, graphics->page_unit, gdip_get_display_dpi (), graphics->type, *height);
 	return Ok;
 }
 
@@ -941,7 +1138,7 @@ GdipGetFontHeightGivenDPI (GDIPCONST GpFont *font, float dpi, float *height)
 	GdipGetEmHeight (font->family, font->style, &emHeight);
 	GdipGetLineSpacing (font->family, font->style, &lineSpacing);
 	*height = lineSpacing * (font->emSize / emHeight);
-	gdip_unit_conversion (font->unit, UnitInch, dpi, gtMemoryBitmap, *height, height);
+	*height = gdip_unit_conversion (font->unit, UnitInch, dpi, gtMemoryBitmap, *height);
 	*height = *height * dpi;
 	return Ok;
 }
