@@ -23,8 +23,7 @@
  *   Duncan Mak (duncan@ximian.com)
  */
 
-#include "gdip.h"
-#include <dlfcn.h>
+#include "general.h"
 #include <cairo.h>
 #include <math.h>
 
@@ -49,17 +48,23 @@ struct startupOutput
 
 
 static GList* g_mem_allocations;
+static BOOL startup = FALSE;
 
 extern void cairo_test_xlib_disable_render();
 
 GpStatus 
 GdiplusStartup(unsigned long *token, const struct startupInput *input, struct startupOutput *output)
 {
-	g_mem_allocations = NULL;
-        initCodecList (); 
-	*token = 1;
-	gdip_get_display_dpi();
-//	cairo_test_xlib_disable_render();
+	/* don't initialize multiple time, e.g. for each appdomain */
+	if (!startup) {
+		startup = TRUE;
+		g_mem_allocations = NULL;
+		initCodecList (); 
+		FcInit ();
+		*token = 1;
+		gdip_get_display_dpi();
+//		cairo_test_xlib_disable_render();
+	}
 	return Ok;
 }
 
@@ -69,8 +74,11 @@ GdiplusShutdown(unsigned long *token)
 #ifdef DEBUG_MEMLEAKS
 	GList* list = NULL;
 #endif
-	releaseCodecList ();
-
+	if (startup) {
+		releaseCodecList ();
+		gdip_font_clear_pattern_cache ();
+		startup = FALSE; /* in case we want to restart it */
+	}
 #ifdef DEBUG_MEMLEAKS
 	for (list = g_list_first (g_mem_allocations); list != NULL; list = g_list_next (list)) {
 		printf ("Memory block not free'd at %x\n", list->data);
@@ -107,6 +115,12 @@ GdipCalloc (size_t nelem, size_t elsize)
 
 }
 
+void *
+GdipRealloc (void *org, int size)
+{
+	return realloc (org, size);
+}
+
 void 
 GdipFree (void *ptr)
 {
@@ -235,6 +249,8 @@ convert_points (const GpPoint *point, int count)
 {
         int i;
         GpPointF *retval = (GpPointF *) GdipAlloc (sizeof (GpPointF) * count);
+	if (!retval)
+		return NULL;
 
         for (i = 0; i < count; i++) {
                 retval [i].X = (float) point [i].X;
@@ -251,6 +267,8 @@ gdip_open_curve_tangents (int terms, const GpPointF *points, int count, float te
         int i;
 
         GpPointF *tangents = (GpPointF *) GdipAlloc (sizeof (GpPointF) * count);
+	if (!tangents)
+		return NULL;
 
         /* initialize everything to zero to begin with */
         for (i = 0; i < count; i++) {
@@ -281,6 +299,8 @@ gdip_closed_curve_tangents (int terms, const GpPointF *points, int count, float
         float coefficient = tension / 3.0;
         int i;
         GpPointF *tangents = (GpPointF *) GdipAlloc (sizeof (GpPointF) * count);
+	if (!tangents)
+		return NULL;
 
         /* initialize everything to zero to begin with */
         for (i = 0; i < count; i++) {
@@ -458,7 +478,7 @@ utf8_to_ucs2(const gchar *utf8, gunichar2 *ucs2, int ucs2_len) {
 	glong		items_read;
 	glong		count;
 	gunichar	*ucs4;
-	unsigned char	*ptr;
+	gunichar2	*ptr;
 
 	items_read = 0;
 	count = 0;
@@ -470,12 +490,11 @@ utf8_to_ucs2(const gchar *utf8, gunichar2 *ucs2, int ucs2_len) {
 		return FALSE;
 	}
 
-	ptr = (unsigned char *)ucs2;
+	ptr = (gunichar2 *)ucs2;
 	for (i = 0; (i < count) && (i < ucs2_len); i++) {
 		if (ucs4[i] < 0x1000 && !(ucs4[i] >= 0xd800 && ucs4[i] < 0xe000)) {
-			ptr[0] = (unsigned char)ucs4[i];
-			ptr[1] = (unsigned char)(ucs4[i] >> 8);
-			ptr += 2;
+			*ptr = (gunichar2)ucs4[i];
+			ptr++;
 		}	/* we're simply ignoring any chars that don't fit into ucs2 */
 	}
 	ucs2[i] = 0;	/* terminate */
@@ -539,3 +558,142 @@ from_cairoformat_to_content (cairo_format_t format)
     }
 }
 
+GpStatus
+gdip_get_pattern_status (cairo_pattern_t *pat)
+{
+	if (!pat)
+		return OutOfMemory;
+
+	switch (cairo_pattern_status (pat)) {
+	case CAIRO_STATUS_SUCCESS:
+		return Ok;
+	case CAIRO_STATUS_NO_MEMORY:
+		cairo_pattern_destroy (pat);
+		return OutOfMemory;
+	default:
+		cairo_pattern_destroy (pat);
+		return InvalidParameter;
+	}
+}
+
+/* cairo has a (signed) 15(1)/16(2)bits pixel positioning, while GDI+ use (signed) 23 bits (infinity).
+ * Using larger values confuse the bits used for subpixel positioning.
+ * (1) http://lists.freedesktop.org/archives/cairo/2006-June/007251.html
+ * (2) testing shows artefacts if using more than 15 *signed* bits (i.e. -16384 to 16383)
+ */
+
+#define CAIRO_LOW_LIMIT		-16384
+#define CAIRO_HIGH_LIMIT	16383
+#define CAIRO_LIMIT(v)		((v < CAIRO_LOW_LIMIT) ? CAIRO_LOW_LIMIT : (v > CAIRO_HIGH_LIMIT) ? CAIRO_HIGH_LIMIT : v)
+
+void
+gdip_cairo_rectangle (GpGraphics *graphics, double x, double y, double width, double height, BOOL antialiasing)
+{
+	double x2, y2;
+
+	/* avoid unit conversion whenever possible */
+	if (!OPTIMIZE_CONVERSION (graphics)) {
+		x = gdip_unitx_convgr (graphics, x);
+		y = gdip_unity_convgr (graphics, y);
+		width = gdip_unitx_convgr (graphics, width);
+		height = gdip_unity_convgr (graphics, height);
+	}
+
+	/* apply antialiasing offset (if required) */
+	if (antialiasing) {
+		x += graphics->aa_offset_x;
+		y += graphics->aa_offset_y;
+	}
+
+	/* ensure that each point (x,y and x+width,y+height) are within the 16 bits bounds */
+	x2 = x + width;
+	y2 = y + height;
+
+	/* put everything between cairo limits */
+	x = CAIRO_LIMIT (x);
+	y = CAIRO_LIMIT (y);
+	x2 = CAIRO_LIMIT (x2);
+	y2 = CAIRO_LIMIT (y2);
+
+	/* and recompute the final width and length */
+	cairo_rectangle (graphics->ct, x, y, (x2 - x), (y2 - y));
+}
+
+void
+gdip_cairo_move_to (GpGraphics *graphics, double x, double y, BOOL convert_units, BOOL antialiasing)
+{
+	/* avoid unit conversion whenever possible */
+	if (convert_units && !OPTIMIZE_CONVERSION (graphics)) {
+		x = gdip_unitx_convgr (graphics, x);
+		y = gdip_unity_convgr (graphics, y);
+	}
+
+	/* apply antialiasing offset (if required) */
+	if (antialiasing) {
+		x += graphics->aa_offset_x;
+		y += graphics->aa_offset_y;
+	}
+
+	/* put everything between cairo limits */
+	x = CAIRO_LIMIT (x);
+	y = CAIRO_LIMIT (y);
+
+        cairo_move_to (graphics->ct, x, y);
+}
+
+void
+gdip_cairo_line_to (GpGraphics *graphics, double x, double y, BOOL convert_units, BOOL antialiasing)
+{
+	/* avoid unit conversion whenever possible */
+	if (convert_units && !OPTIMIZE_CONVERSION (graphics)) {
+		x = gdip_unitx_convgr (graphics, x);
+		y = gdip_unity_convgr (graphics, y);
+	}
+
+	/* apply antialiasing offset (if required) */
+	if (antialiasing) {
+		x += graphics->aa_offset_x;
+		y += graphics->aa_offset_y;
+	}
+
+	/* put everything between cairo limits */
+	x = CAIRO_LIMIT (x);
+	y = CAIRO_LIMIT (y);
+
+        cairo_line_to (graphics->ct, x, y);
+}
+
+void
+gdip_cairo_curve_to (GpGraphics *graphics, double x1, double y1, double x2, double y2, double x3, double y3, 
+	BOOL convert_units, BOOL antialiasing)
+{
+	/* avoid unit conversion whenever possible */
+	if (convert_units && !OPTIMIZE_CONVERSION (graphics)) {
+		x1 = gdip_unitx_convgr (graphics, x1);
+		y1 = gdip_unity_convgr (graphics, y1);
+		x2 = gdip_unitx_convgr (graphics, x2);
+		y2 = gdip_unity_convgr (graphics, y2);
+		x3 = gdip_unitx_convgr (graphics, x3);
+		y3 = gdip_unity_convgr (graphics, y3);
+	}
+
+	/* apply antialiasing offset (if required) */
+	if (antialiasing) {
+		x1 += graphics->aa_offset_x;
+		y1 += graphics->aa_offset_y;
+		x2 += graphics->aa_offset_x;
+		y2 += graphics->aa_offset_y;
+		x3 += graphics->aa_offset_x;
+		y3 += graphics->aa_offset_y;
+	}
+
+	/* put everything between cairo limits */
+	x1 = CAIRO_LIMIT (x1);
+	y1 = CAIRO_LIMIT (y1);
+	x2 = CAIRO_LIMIT (x2);
+	y2 = CAIRO_LIMIT (y2);
+	x3 = CAIRO_LIMIT (x3);
+	y3 = CAIRO_LIMIT (y3);
+
+	cairo_curve_to (graphics->ct, x1, y1, x2, y2, x3, y3);
+}
